@@ -1,7 +1,7 @@
 (*
  * Copyright (c) 1997-1999 Massachusetts Institute of Technology
- * Copyright (c) 2003 Matteo Frigo
- * Copyright (c) 2003 Massachusetts Institute of Technology
+ * Copyright (c) 2003, 2007-11 Matteo Frigo
+ * Copyright (c) 2003, 2007-11 Massachusetts Institute of Technology
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,16 +15,14 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  *
  *)
-(* $Id: gen_notw.ml,v 1.22 2003/04/17 11:07:19 athena Exp $ *)
 
 open Util
 open Genutil
 open C
 
-let cvsid = "$Id: gen_notw.ml,v 1.22 2003/04/17 11:07:19 athena Exp $"
 
 let usage = "Usage: " ^ Sys.argv.(0) ^ " -n <number>"
 
@@ -51,6 +49,12 @@ let speclist = [
   " specialize for given output vector stride"
 ] 
 
+let nonstandard_optimizer list_of_buddy_stores dag =
+  let sched = standard_scheduler dag in
+  let annot = Annotate.annotate list_of_buddy_stores sched in
+  let _ = dump_asched annot in
+  annot
+
 let generate n =
   let riarray = "ri"
   and iiarray = "ii"
@@ -64,64 +68,79 @@ let generate n =
 
   let sign = !Genutil.sign 
   and name = !Magic.codelet_name
-  and byvl x = choose_simd x (ctimes (CVar "VL", x))  in
+  and byvl x = choose_simd x (ctimes (CVar "(2 * VL)", x))  in
   let ename = expand_name name in
 
-  let vl = choose_simd "1" "VL"
-  in
-
   let vistride = either_stride (!uistride) (C.SVar istride)
   and vostride = either_stride (!uostride) (C.SVar ostride)
   in
 
-  let _ = Simd.ovs := stride_to_string "ovs" !uovstride in
-  let _ = Simd.ivs := stride_to_string "ivs" !uivstride in
+  let sovs = stride_to_string "ovs" !uovstride in
+  let sivs = stride_to_string "ivs" !uivstride in
 
   let locations = unique_array_c n in
   let input = 
     locative_array_c n 
       (C.array_subscript riarray vistride)
       (C.array_subscript iiarray vistride)
-      locations in
+      locations sivs in
   let output = Fft.dft sign n (load_array_c n input) in
   let oloc = 
     locative_array_c n 
       (C.array_subscript roarray vostride)
       (C.array_subscript ioarray vostride)
-      locations in
+      locations sovs in
+  let list_of_buddy_stores =
+    let k = !Simdmagic.store_multiple in
+    if (k > 1) then
+      if (n mod k == 0) then
+	List.append
+	  (List.map 
+	     (fun i -> List.map (fun j -> (fst (oloc (k * i + j)))) (iota k))
+	     (iota (n / k)))
+	  (List.map 
+	     (fun i -> List.map (fun j -> (snd (oloc (k * i + j)))) (iota k))
+	     (iota (n / k)))
+      else failwith "invalid n for -store-multiple"
+    else []
+  in
+
   let odag = store_array_c n oloc output in
-  let annot = standard_optimizer odag in
+  let annot = nonstandard_optimizer list_of_buddy_stores odag in
 
   let body = Block (
-    [Decl ("int", i)],
+    [Decl ("INT", i)],
     [For (Expr_assign (CVar i, CVar v),
 	  Binop (" > ", CVar i, Integer 0),
 	  list_to_comma 
 	    [Expr_assign (CVar i, CPlus [CVar i; CUminus (byvl (Integer 1))]);
 	     Expr_assign (CVar riarray, CPlus [CVar riarray; 
-					       byvl (CVar !Simd.ivs)]);
+					       byvl (CVar sivs)]);
 	     Expr_assign (CVar iiarray, CPlus [CVar iiarray; 
-					       byvl (CVar !Simd.ivs)]);
+					       byvl (CVar sivs)]);
 	     Expr_assign (CVar roarray, CPlus [CVar roarray; 
-					       byvl (CVar !Simd.ovs)]);
+					       byvl (CVar sovs)]);
 	     Expr_assign (CVar ioarray, CPlus [CVar ioarray; 
-					       byvl (CVar !Simd.ovs)])],
+					       byvl (CVar sovs)]);
+	     make_volatile_stride (CVar istride);
+	     make_volatile_stride (CVar ostride)
+	   ],
 	  Asch annot)
    ])
   in
 
   let tree =
-    Fcn ("static void", ename,
+    Fcn ((if !Magic.standalone then "void" else "static void"), ename,
 	 ([Decl (C.constrealtypep, riarray);
 	   Decl (C.constrealtypep, iiarray);
 	   Decl (C.realtypep, roarray);
  	   Decl (C.realtypep, ioarray);
 	   Decl (C.stridetype, istride);
 	   Decl (C.stridetype, ostride);
-	   Decl ("int", v);
-	   Decl ("int", "ivs");
-	   Decl ("int", "ovs")]),
-	 add_constants body)
+	   Decl ("INT", v);
+	   Decl ("INT", "ivs");
+	   Decl ("INT", "ovs")]),
+	 finalize_fcn body)
 
   in let desc = 
     Printf.sprintf 
@@ -137,9 +156,8 @@ let generate n =
     "  X(kdft_register)(p, " ^ ename ^ ", &desc);\n" ^
     "}\n"
 
-  in ((unparse cvsid tree) ^ "\n" ^ 
-      desc ^
-      init)
+  in ((unparse tree) ^ "\n" ^ 
+      (if !Magic.standalone then "" else desc ^ init))
 
 let main () =
   begin
