@@ -1,7 +1,7 @@
 (*
  * Copyright (c) 1997-1999 Massachusetts Institute of Technology
- * Copyright (c) 2003 Matteo Frigo
- * Copyright (c) 2003 Massachusetts Institute of Technology
+ * Copyright (c) 2003, 2007-11 Matteo Frigo
+ * Copyright (c) 2003, 2007-11 Massachusetts Institute of Technology
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,10 +15,9 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  *
  *)
-(* $Id: gen_r2r.ml,v 1.3 2003/04/17 19:25:50 athena Exp $ *)
 
 (* generation of trigonometric transforms *)
 
@@ -26,7 +25,6 @@ open Util
 open Genutil
 open C
 
-let cvsid = "$Id: gen_r2r.ml,v 1.3 2003/04/17 19:25:50 athena Exp $"
 
 let usage = "Usage: " ^ Sys.argv.(0) ^ " -n <number>"
 
@@ -50,6 +48,9 @@ type mode =
   | NONE
 
 let mode = ref NONE
+let normsqr = ref 1
+let unitary = ref false
+let noloop = ref false
 
 let speclist = [
   "-with-istride",
@@ -111,98 +112,118 @@ let speclist = [
   "-rodft11",
   Arg.Unit(fun () -> mode := RODFT11),
   " generate a DST-IV codelet";
+
+  "-normalization",
+  Arg.String(fun x -> let ix = int_of_string x in normsqr := ix * ix),
+  " normalization integer to divide by";
+
+  "-normsqr",
+  Arg.String(fun x -> normsqr := int_of_string x),
+  " integer square of normalization to divide by";
+
+  "-unitary",
+  Arg.Unit(fun () -> unitary := true),
+  " unitary normalization (up overall scale factor)";
+
+  "-noloop",
+  Arg.Unit(fun () -> noloop := true),
+  " no vector loop";
 ]
 
+let sqrt_half = Complex.inverse_int_sqrt 2
+let sqrt_two = Complex.int_sqrt 2
+
+let rescale sc s1 s2 input i = 
+  if ((i == s1 || i == s2) && !unitary) then
+    Complex.times (input i) sc
+  else
+    input i
+
 let generate n mode =
   let iarray = "I"
   and oarray = "O"
-  and istride = "istride"
-  and ostride = "ostride" in
+  and istride = "is"
+  and ostride = "os" 
+  and i = "i" 
+  and v = "v" 
+  in
 
-  let ns = string_of_int n
-  and sign = !Genutil.sign 
+  let sign = !Genutil.sign 
   and name = !Magic.codelet_name in
-  let name0 = name ^ "_0" in
 
   let vistride = either_stride (!uistride) (C.SVar istride)
   and vostride = either_stride (!uostride) (C.SVar ostride)
   in
 
-  let _ = Simd.ovs := stride_to_string "ovs" !uovstride in
-  let _ = Simd.ivs := stride_to_string "ivs" !uivstride in
+  let sovs = stride_to_string "ovs" !uovstride in
+  let sivs = stride_to_string "ivs" !uivstride in
 
-  let (transform, load_input, store_output) = match mode with
-  | RDFT -> Trig.rdft sign, load_array_r, store_array_hc
-  | HDFT -> Trig.hdft sign, load_array_c, store_array_r  (* TODO *)
-  | DHT -> Trig.dht 1, load_array_r, store_array_r
-  | REDFT00 -> Trig.dctI, load_array_r, store_array_r
-  | REDFT10 -> Trig.dctII, load_array_r, store_array_r
-  | REDFT01 -> Trig.dctIII, load_array_r, store_array_r
-  | REDFT11 -> Trig.dctIV, load_array_r, store_array_r
-  | RODFT00 -> Trig.dstI, load_array_r, store_array_r
-  | RODFT10 -> Trig.dstII, load_array_r, store_array_r
-  | RODFT01 -> Trig.dstIII, load_array_r, store_array_r
-  | RODFT11 -> Trig.dstIV, load_array_r, store_array_r
+  let (transform, load_input, store_output, si1,si2,so1,so2) = match mode with
+  | RDFT -> Trig.rdft sign, load_array_r, store_array_hc, -1,-1,-1,-1
+  | HDFT -> Trig.hdft sign, load_array_c, store_array_r, -1,-1,-1,-1 (* TODO *)
+  | DHT -> Trig.dht 1, load_array_r, store_array_r, -1,-1,-1,-1
+  | REDFT00 -> Trig.dctI, load_array_r, store_array_r, 0,n-1,0,n-1
+  | REDFT10 -> Trig.dctII, load_array_r, store_array_r, -1,-1,0,-1
+  | REDFT01 -> Trig.dctIII, load_array_r, store_array_r, 0,-1,-1,-1
+  | REDFT11 -> Trig.dctIV, load_array_r, store_array_r, -1,-1,-1,-1
+  | RODFT00 -> Trig.dstI, load_array_r, store_array_r, -1,-1,-1,-1
+  | RODFT10 -> Trig.dstII, load_array_r, store_array_r, -1,-1,n-1,-1
+  | RODFT01 -> Trig.dstIII, load_array_r, store_array_r, n-1,-1,-1,-1
+  | RODFT11 -> Trig.dstIV, load_array_r, store_array_r, -1,-1,-1,-1
   | _ -> failwith "must specify transform kind"
   in
     
   let locations = unique_array_c n in
-  let input = 
-    locative_array_c n 
+  let input = locative_array_c n 
       (C.array_subscript iarray vistride)
       (C.array_subscript "BUG" vistride)
-      locations in
-  let output = transform n (load_array_c n input) in
+      locations sivs in
+  let output = rescale sqrt_half so1 so2
+      ((Complex.times (Complex.inverse_int_sqrt !normsqr))
+       @@ (transform n (rescale sqrt_two si1 si2 (load_array_c n input)))) in
   let oloc = 
     locative_array_c n 
       (C.array_subscript oarray vostride)
       (C.array_subscript "BUG" vostride)
-      locations in
+      locations sovs in
   let odag = store_output n oloc output in
   let annot = standard_optimizer odag in
 
-  let tree0 =
-    Fcn ("static void", name0,
+  let body = if !noloop then Block([], [Asch annot]) else Block (
+    [Decl ("INT", i)],
+    [For (Expr_assign (CVar i, CVar v),
+	  Binop (" > ", CVar i, Integer 0),
+	  list_to_comma 
+	    [Expr_assign (CVar i, CPlus [CVar i; CUminus (Integer 1)]);
+	     Expr_assign (CVar iarray, CPlus [CVar iarray; CVar sivs]);
+	     Expr_assign (CVar oarray, CPlus [CVar oarray; CVar sovs]);
+	     make_volatile_stride (CVar istride);
+	     make_volatile_stride (CVar ostride)
+	   ],
+	  Asch annot)
+   ])
+  in
+
+  let tree =
+    Fcn ((if !Magic.standalone then "void" else "static void"), name,
 	 ([Decl (C.constrealtypep, iarray);
 	   Decl (C.realtypep, oarray)]
 	  @ (if stride_fixed !uistride then [] 
                else [Decl (C.stridetype, istride)])
 	  @ (if stride_fixed !uostride then [] 
 	       else [Decl (C.stridetype, ostride)])
-	  @ (choose_simd []
-	       (if stride_fixed !uivstride then [] else 
-	       [Decl ("int", !Simd.ivs)]))
-	  @ (choose_simd []
-	       (if stride_fixed !uovstride then [] else 
-	       [Decl ("int", !Simd.ovs)]))
-	 ),
-	 add_constants (Asch annot))
+	  @ (if !noloop then [] else
+               [Decl ("INT", v)]
+	       @ (if stride_fixed !uivstride then [] 
+                    else [Decl ("INT", "ivs")])
+	       @ (if stride_fixed !uovstride then [] 
+                    else [Decl ("INT", "ovs")]))),
+	 finalize_fcn body)
 
-  in let loop =
-    "static void " ^ name ^
-    "(const " ^ C.realtype ^ " *I, " ^ 
-    C.realtype ^ " *O, " ^
-    C.stridetype ^ " is, " ^ 
-    C.stridetype ^ " os, " ^ 
-      " int v, int ivs, int ovs)\n" ^
-    "{\n" ^
-    "int i;\n" ^
-    "for (i = v; i > 0; --i) {\n" ^
-      name0 ^ "(I, O" ^
-       (if stride_fixed !uistride then "" else ", is") ^ 
-       (if stride_fixed !uostride then "" else ", os") ^ 
-       (choose_simd ""
-	  (if stride_fixed !uivstride then "" else ", ivs")) ^ 
-       (choose_simd ""
-	  (if stride_fixed !uovstride then "" else ", ovs")) ^ 
-    ");\n" ^
-    "I += ivs; O += ovs;\n" ^
-    "}\n}\n\n"
-
-  and desc = 
+  in let desc = 
     Printf.sprintf 
-      "static const kr2r_desc desc = { %d, \"%s\", %s, &GENUS, %s, %s, %s, %s, %s };\n\n"
-      n name (flops_of tree0) 
+      "static const kr2r_desc desc = { %d, \"%s\", %s, &GENUS, %s };\n\n"
+      n name (flops_of tree) 
       (match mode with
       | RDFT -> "RDFT00"
       | HDFT -> "HDFT00"
@@ -216,10 +237,6 @@ let generate n mode =
       | RODFT01 -> "RODFT01"
       | RODFT11 -> "RODFT11"
       | _ -> failwith "must specify a transform kind")
-      (stride_to_solverparm !uistride) 
-      (stride_to_solverparm !uostride)
-      (choose_simd "0" (stride_to_solverparm !uivstride))
-      (choose_simd "0" (stride_to_solverparm !uovstride))
 
   and init =
     (declare_register_fcn name) ^
@@ -228,7 +245,7 @@ let generate n mode =
     "}\n"
 
   in
-  (unparse cvsid tree0) ^ "\n" ^ loop ^ desc ^ init
+  (unparse tree) ^ "\n" ^ (if !Magic.standalone then "" else desc ^ init)
 
 
 let main () =
