@@ -1,7 +1,7 @@
 (*
  * Copyright (c) 1997-1999 Massachusetts Institute of Technology
- * Copyright (c) 2003, 2006 Matteo Frigo
- * Copyright (c) 2003, 2006 Massachusetts Institute of Technology
+ * Copyright (c) 2003, 2007-11 Matteo Frigo
+ * Copyright (c) 2003, 2007-11 Massachusetts Institute of Technology
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,23 +15,21 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  *
  *)
-(* $Id: gen_twiddle.ml,v 1.24 2006-02-12 23:34:12 athena Exp $ *)
 
 open Util
 open Genutil
 open C
 
-let cvsid = "$Id: gen_twiddle.ml,v 1.24 2006-02-12 23:34:12 athena Exp $"
 
 type ditdif = DIT | DIF
 let ditdif = ref DIT
 let usage = "Usage: " ^ Sys.argv.(0) ^ " -n <number> [ -dit | -dif ]"
 
-let uiostride = ref Stride_variable
-let udist = ref Stride_variable
+let urs = ref Stride_variable
+let ums = ref Stride_variable
 
 let speclist = [
   "-dit",
@@ -42,49 +40,46 @@ let speclist = [
   Arg.Unit(fun () -> ditdif := DIF),
   " generate a DIF codelet";
 
-  "-with-iostride",
-  Arg.String(fun x -> uiostride := arg_to_stride x),
+  "-with-rs",
+  Arg.String(fun x -> urs := arg_to_stride x),
   " specialize for given i/o stride";
 
-  "-with-dist",
-  Arg.String(fun x -> udist := arg_to_stride x),
-  " specialize for given dist"
+  "-with-ms",
+  Arg.String(fun x -> ums := arg_to_stride x),
+  " specialize for given ms"
 ]
 
 let generate n =
   let rioarray = "ri"
   and iioarray = "ii"
-  and iostride = "ios"
+  and rs = "rs"
   and twarray = "W"
-  and i = "i" 
-  and m = "m"
-  and dist = "dist" in
+  and m = "m" and mb = "mb" and me = "me" and ms = "ms" in
 
   let sign = !Genutil.sign 
   and name = !Magic.codelet_name 
   and byvl x = choose_simd x (ctimes (CVar "(2 * VL)", x)) in
   let ename = expand_name name in
 
-  let (bytwiddle, num_twiddles, twdesc) = Twiddle.twiddle_policy false in
+  let (bytwiddle, num_twiddles, twdesc) = Twiddle.twiddle_policy 0 false in
   let nt = num_twiddles n in
 
   let byw = bytwiddle n sign (twiddle_array nt twarray) in
 
-  let viostride = either_stride (!uiostride) (C.SVar iostride) in
-  let _ = Simd.ovs := stride_to_string "dist" !udist in
-  let _ = Simd.ivs := stride_to_string "dist" !udist in
+  let vrs = either_stride (!urs) (C.SVar rs) in
+  let sms = stride_to_string "ms" !ums in
 
   let locations = unique_array_c n in
   let iloc = 
     locative_array_c n 
-      (C.array_subscript rioarray viostride)
-      (C.array_subscript iioarray viostride)
-      locations
+      (C.array_subscript rioarray vrs)
+      (C.array_subscript iioarray vrs)
+      locations sms
   and oloc = 
     locative_array_c n 
-      (C.array_subscript rioarray viostride)
-      (C.array_subscript iioarray viostride)
-      locations 
+      (C.array_subscript rioarray vrs)
+      (C.array_subscript iioarray vrs)
+      locations sms
   in
   let liloc = load_array_c n iloc in
   let output =
@@ -95,44 +90,50 @@ let generate n =
   let odag = store_array_c n oloc output in
   let annot = standard_optimizer odag in
 
+  let vm = CVar m and vmb = CVar mb and vme = CVar me in
+
   let body = Block (
-    [Decl ("INT", i)],
-    [For (Expr_assign (CVar i, CVar m),
-	  Binop (" > ", CVar i, Integer 0),
+    [Decl ("INT", m)],
+    [For (list_to_comma
+	    [Expr_assign (vm, vmb);
+	     Expr_assign (CVar twarray, 
+			  CPlus [CVar twarray; 
+				 ctimes (vmb, Integer nt)])],
+	  Binop (" < ", vm, vme),
 	  list_to_comma 
-	    [Expr_assign (CVar i, CPlus [CVar i; CUminus (byvl (Integer 1))]);
+	    [Expr_assign (vm, CPlus [vm; byvl (Integer 1)]);
 	     Expr_assign (CVar rioarray, CPlus [CVar rioarray; 
-						byvl (CVar !Simd.ivs)]);
+						byvl (CVar sms)]);
 	     Expr_assign (CVar iioarray, CPlus [CVar iioarray; 
-						byvl (CVar !Simd.ivs)]);
+						byvl (CVar sms)]);
 	     Expr_assign (CVar twarray, CPlus [CVar twarray; 
 					       byvl (Integer nt)]);
-	     make_volatile_stride (CVar iostride)
-	   ],
-	  Asch annot);
-     Return (CVar twarray)
-   ])
+	     make_volatile_stride (CVar rs)
+	    ],
+	  Asch annot)])
   in
 
   let tree = 
-    Fcn (("static " ^ C.constrealtypep), ename,
+    Fcn (((if !Magic.standalone then "" else "static ") ^ "void"),
+	 ename,
 	 [Decl (C.realtypep, rioarray);
 	  Decl (C.realtypep, iioarray);
 	  Decl (C.constrealtypep, twarray);
-	  Decl (C.stridetype, iostride);
-	  Decl ("INT", m);
-	  Decl ("INT", dist)],
-         add_constants body)
+	  Decl (C.stridetype, rs);
+	  Decl ("INT", mb);
+	  Decl ("INT", me);
+	  Decl ("INT", ms)],
+         finalize_fcn body)
   in
   let twinstr = 
     Printf.sprintf "static const tw_instr twinstr[] = %s;\n\n" 
       (twinstr_to_string "(2 * VL)" (twdesc n))
   and desc = 
     Printf.sprintf
-      "static const ct_desc desc = {%d, \"%s\", twinstr, &GENUS, %s, %s, %s, %s};\n\n"
-      n name (flops_of tree) 
-      (stride_to_solverparm !uiostride) "0"
-      (stride_to_solverparm !udist) 
+      "static const ct_desc desc = {%d, %s, twinstr, &GENUS, %s, %s, %s, %s};\n\n"
+      n (stringify name) (flops_of tree) 
+      (stride_to_solverparm !urs) "0"
+      (stride_to_solverparm !ums) 
   and register = 
     match !ditdif with
     | DIT -> "X(kdft_dit_register)"
@@ -147,7 +148,8 @@ let generate n =
     (Printf.sprintf "{\n%s(p, %s, &desc);\n}" register ename)
   in
 
-  (unparse cvsid tree) ^ "\n" ^ init
+  (unparse tree) ^ "\n" ^
+    (if !Magic.standalone then "" else init)
 
 
 let main () =
